
import numpy as np
from nbtlib.tag import Int, Long, Double, String, List, Compound, IntArray
from math import ceil, log



from .info import *

from .storage import LitematicaBitArray, DiscriminatingDictionary



class Region:
    """
    A schematic region
    x, y, z: position in the schematic (read only)
    width, height, length: size of the region (oriented, can be negative)
    """

    def __init__(self, x, y, z, width, height, length):
        from .Schematic import AIR
        
        if width == 0 or height == 0 or length == 0:
            raise ValueError("Region dimensions cannot be 0")
        self.__x, self.__y, self.__z = x, y, z
        self.__width, self.__height, self.__length = width, height, length
        self.__palette = [AIR, ]
        self.__blocks = np.zeros((abs(width), abs(height), abs(length)), dtype=np.uint32)
        self.entities = []
        self.block_entities = []
        self.blockTicks = []
        self.fluidTicks = []

    def _tonbt(self):
        """
        Write this region to an nbt tab and return it
        """
        root = Compound()
        pos = Compound()
        pos["x"] = Int(self.__x)
        pos["y"] = Int(self.__y)
        pos["z"] = Int(self.__z)
        root["Position"] = pos
        size = Compound()
        size["x"] = Int(self.__width)
        size["y"] = Int(self.__height)
        size["z"] = Int(self.__length)
        root["Size"] = size

        plt = List[Compound]([blk._tonbt() for blk in self.__palette])
        root["BlockStatePalette"] = plt

        entities = List[Compound]([entity._tonbt() for entity in self.entities])
        root["Entities"] = entities

        block_entities = List[Compound]([block_entity._tonbt() for block_entity in self.block_entities])
        root["TileEntities"] = block_entities

        root["PendingBlockTicks"] = List[Compound](self.blockTicks) #TODO How does this work
        root["PendingFluidTicks"] = List[Compound](self.fluidTicks)

        arr = LitematicaBitArray(self.getvolume(), self.__get_needed_nbits())
        for x in range(abs(self.__width)):
            for y in range(abs(self.__height)):
                for z in range(abs(self.__length)):
                    ind = (y * abs(self.__width * self.__length)) + z * abs(self.__width) + x
                    arr[ind] = int(self.__blocks[x, y, z])
        root["BlockStates"] = arr._tonbtlongarray()

        return root

    def getblock(self, x, y, z):
        """
        Return the Block(State) at the given coordinates
        """
        x, y, z = self.__regcoordinates2storecoords(x, y, z)
        return self.__palette[self.__blocks[x, y, z]]


    def setblock(self, x, y, z, block):
        """
        Set the Block(State) at the given coordinate
        """

        x, y, z = self.__regcoordinates2storecoords(x, y, z)
        if block in self.__palette:
            i = self.__palette.index(block)
        else:
            self.__palette.append(block)
            i = len(self.__palette) - 1
        self.__blocks[x, y, z] = i

    def getblockcount(self):
        """
        Returns the number of non-air in the region
        """
        airind = self.__palette.index()
        c = 0
        for block in self.__blocks.flat:
            if block != airind:
                c += 1
        return c

    def __regcoordinates2storecoords(self, x, y, z):
        if self.__width < 0:
            x -= self.__width + 1
        if self.__height < 0:
            y -= self.__height + 1
        if self.__length < 0:
            z -= self.__length + 1
        return x, y, z

    def getvolume(self):
        """
        Returns the region's volume
        """
        return abs(self.__width * self.__height * self.__length)

    def __get_needed_nbits(self):
        return max(ceil(log(len(self.__palette), 2)), 2)

    @staticmethod
    def fromnbt(nbt):
        """
        Read a region from an nbt tag and return it
        """

        from .BlockState import BlockState
        from .BlockEntity import BlockEntity
        from .Entity import Entity

        pos = nbt["Position"]
        x = int(pos["x"])
        y = int(pos["y"])
        z = int(pos["z"])
        size = nbt["Size"]
        width = int(size["x"])
        height = int(size["y"])
        length = int(size["z"])
        reg = Region(x, y, z, width, height, length)
        del reg.__palette[0]
        for bnbt in nbt["BlockStatePalette"]:
            block = BlockState.fromnbt(bnbt)
            reg.__palette.append(block)

        for entity_nbt in nbt["Entities"]:
            entity = Entity.fromnbt(entity_nbt)
            reg.entities.append(entity)

        for block_entity_nbt in nbt["TileEntities"]:
            block = BlockEntity.fromnbt(block_entity_nbt)
            reg.block_entities.append(block)

        blks = nbt["BlockStates"]
        nbits = reg.__get_needed_nbits()
        arr = LitematicaBitArray.fromnbtlongarray(blks, reg.getvolume(), nbits)
        for x in range(abs(width)):
            for y in range(abs(height)):
                for z in range(abs(length)):
                    ind = (y * abs(width * length)) + z * abs(width) + x
                    reg.__blocks[x][y][z] = arr[ind]

        for blockTick in nbt["PendingBlockTicks"]:
            reg.blockTicks.append(blockTick)

        for fluidTick in nbt["PendingFluidTicks"]:
            reg.fluidTicks.append(fluidTick)

        return reg

    def minschemx(self):
        """
        Returns the minimum X coordinate of this region in the schematics coordinate system
        """
        return min(self.__x, self.__x + self.width + 1)

    def maxschemx(self):
        """
        Returns the maximum X coordinate of this region in the schematics coordinate system
        """
        return max(self.__x, self.__x + self.width - 1)

    def minschemy(self):
        """
        Returns the minimum Y coordinate of this region in the schematics coordinate system
        """
        return min(self.__y, self.__y + self.height + 1)

    def maxschemy(self):
        """
        Returns the maximum Y coordinate of this region in the schematics coordinate system
        """
        return max(self.__y, self.__y + self.height - 1)

    def minschemz(self):
        """
        Returns the minimum Z coordinate of this region in the schematics coordinate system
        """
        return min(self.__z, self.__z + self.length + 1)

    def maxschemz(self):
        """
        Returns the maximum Z coordinate of this region in the schematics coordinate system
        """
        return max(self.__z, self.__z + self.length - 1)

    def minx(self):
        """
        Returns the minimum X coordinate of this region in its own coordinate system
        """
        return min(0, self.width + 1)

    def maxx(self):
        """
        Returns the maximum X coordinate of this region in its own coordinate system
        """
        return max(0, self.width - 1)

    def miny(self):
        """
        Returns the minimum Y coordinate of this region in its own coordinate system
        """
        return min(0, self.height + 1)

    def maxy(self):
        """
        Returns the maximum Y coordinate of this region in its own coordinate system
        """
        return max(0, self.height - 1)

    def minz(self):
        """
        Returns the minimum Z coordinate of this region in its own coordinate system
        """
        return min(0, self.length + 1)

    def maxz(self):
        """
        Returns the maximum Z coordinate of this region in its own coordinate system
        """
        return max(0, self.length - 1)

    def xrange(self):
        """
        Returns the range of coordinates this region contains along it's X axis
        """
        return range(self.minx(), self.maxx() + 1)

    def yrange(self):
        """
        Returns the range of coordinates this region contains along it's Y axis
        """
        return range(self.miny(), self.maxy() + 1)

    def zrange(self):
        """
        Returns the range of coordinates this region contains along it's Z axis
        """
        return range(self.minz(), self.maxz() + 1)

    def allblockpos(self):
        """
        Returns an iterator over the coordinates this regions contains in its own coordinate system
        """
        for x in self.xrange():
            for y in self.yrange():
                for z in self.zrange():
                    yield x, y, z

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    @property
    def z(self):
        return self.__z

    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @property
    def length(self):
        return self.__length


    def SetOrigin(self,x,y,z):
        """
        Be careful with this. it may break something...
        But never the less, it sets the origin / Pos of the region
        """
        self.__x, self.__y, self.__z = x, y, z

    def as_schematic(self, name=DEFAULT_NAME, author="", description=""):
        """
        Creates and returns a schematic that contains that region at the origin.
        name: A name for both the region and the schematic
        author: an author for the schematic
        description: a description for the schematic
        """
        from .Schematic import Schematic

        return Schematic(name=name, author=author, description=description, regions={name: self})
